import {world,system, Location,BlockLocation,BlockAreaSize,MinecraftBlockTypes,Vector,ItemStack,MinecraftItemTypes,ItemTypes, MinecraftEffectTypes} from "@minecraft/server"
import {BannerColors, BannerPatternIndex} from "./enumerations.js";
import {fromDynamicProperty,editDynamicProperty} from "./libs/ui-from-dialogue/utils.js"
import "./screens/display.js";
import "./screens/variables.js";



function setTimeOut(callback,ticks){
	let params = {ticks: 0};
	let s = world.events.tick.subscribe((e)=>{
		params.ticks+=1;
		if(params.ticks === ticks){
			callback();
			world.events.tick.unsubscribe(s);
		}
	});
}



class CustomShields{
	constructor(names){
		this.shield_names = names;
		this.area = new BlockAreaSize(4,4,4);
		this.selectedSlots = new Map();
		this.resetSelectedSlot();
		this.onSelect();
		this.onHit();
		this.onProjectileHit();
		this.netheriteMaterial(["add:netherite_shield"]);
		this.interactArmorStandListener();
	}
	onProjectileHit(){
		world.events.projectileHit.subscribe((e)=>{
			try{
			let {projectile,hitVector,entityHit} = e;
			if(entityHit?.entity?.typeId!="minecraft:player") return;
			const context = this.getContext(entityHit.entity);
			if(!context.charged) return;
			const targets = entityHit.entity.dimension.getEntities({location: new Location(entityHit.entity.location.x + Math.round(entityHit.entity.viewDirection.x),entityHit.entity.location.y + Math.round(entityHit.entity.viewDirection.y),entityHit.entity.location.z + Math.round(entityHit.entity.viewDirection.z)), volume: this.area, excludeNames: [entityHit.entity.name ]});
            if(!Array.from(targets).some((el)=>el==projectile)) return;
			entityHit.entity.addTag("invalid_projectile");
			
			projectile.setVelocity({ x:-hitVector.x, y:-hitVector.y, z:-hitVector.z});
			this.reduceDurability(context.shield,entityHit.entity);
			this.removeEffectAttacks(entityHit.entity);
			setTimeOut(()=>{ try{entityHit.entity.removeTag("invalid_projectile")}catch(error){}; },10);
			}catch(error){};
		});
	}
	onHit(){
		world.events.entityHit.subscribe(({hitEntity,entity})=>{
			try{
			if(!entity || (hitEntity?.typeId!="minecraft:player")) return;
			const context = this.getContext(hitEntity);
			if(!context.charged) return;
			
			const targets = hitEntity.dimension.getEntities({location: new Location(hitEntity.location.x + Math.round(hitEntity.viewDirection.x),hitEntity.location.y + Math.round(hitEntity.viewDirection.y),hitEntity.location.z + Math.round(hitEntity.viewDirection.z)), volume: this.area, excludeNames: [hitEntity.name ]});
            if(!Array.from(targets).some((el)=>el==entity)) return;
			entity.addTag("damager");
			entity.setVelocity( { x: -(entity.viewDirection.x*1.2), y: entity.viewDirection.y + 0.2, z: -(entity.viewDirection.z*1.2)} );
			this.reduceDurability(context.shield,hitEntity);
			setTimeOut(()=>{ try{ entity.removeTag("damager")}catch(error){}; },10);
			this.removeEffectAttacks(hitEntity);
			} catch (error){ console.warn(error);}
		});
	}
	getContext(player){
		let res = { has: false, charged: false, shield: undefined};
		let item = player.getComponent("minecraft:inventory").container.getItem(player.selectedSlot);
		if(this.shield_names.some((el)=>el==item?.typeId)){
			res.has = true; res.charged = ((player.isSneaking)&&(!player.hasTag("has_vanilla_shield"))); res.shield = item;
		
		}
		(res.charged) ? [player.triggerEvent("minecraft:shield_protection"),player.triggerEvent("minecraft:vanilla_shield")] : [player.triggerEvent("minecraft:remove_shield_protection"),player.triggerEvent("minecraft:vanilla_shield")];
		return res;
	}
	onSelect(){
		world.events.tick.subscribe((e)=>{
			for (const player of world.getPlayers()){
			    const context = this.getContext(player);
				const selected_slot = this.selectedSlots.get(player);
				if(selected_slot != player.selectedSlot){
					
					this.selectedSlots.set(player,player.selectedSlot);
					this.printShieldPatterns(context.shield,player);
				}
			}
		});
	}
	reduceDurability(itemStack,player){
		let component = itemStack.getComponent("minecraft:durability");
		if(!component) return;
		let dmg = component.damageRange.next();
		component.damage = component.damage + 10;
		if(component.damage > component.maxDurability) { player.getComponent("minecraft:inventory").container.getSlot(player.selectedSlot).clearItem(); }else{ player.getComponent("minecraft:inventory").container.setItem(player.selectedSlot,itemStack);}
	}
	async printShieldPatterns(shield,player){
		await player.runCommandAsync("playanimation @s animation.spartan_shield.reset_layers");
		if(!shield) return;
		let layers = shield.getLore();
		player.shieldLayers = layers;
		for(let layer = 0; layer<layers.length; layer++){
			let [subvariant,...variant] = layers[layer].split(" ");
			variant = variant.join("_");
			let type = 0;
			(variant=="banner") ? type = BannerColors[subvariant]+1 : type = ((BannerPatternIndex[variant]*(Object.keys(BannerColors).length))+BannerColors[subvariant]) + 1;
			await player.runCommandAsync(`playanimation @s animation.spartan_shield.layer_${layer}_index_${type}`);
		}
	}
	async printPatternsOnArmorStand(layers,entity){
		await entity.runCommandAsync("playanimation @s animation.spartan_shield.reset_layers");
		for(let layer = 0; layer<layers.length; layer++){
			let [subvariant,...variant] = layers[layer].split(" ");
			variant = variant.join("_");
			let type = 0;
			(variant=="banner") ? type = BannerColors[subvariant]+1 : type = ((BannerPatternIndex[variant]*(Object.keys(BannerColors).length))+BannerColors[subvariant]) + 1;
			await entity.runCommandAsync(`playanimation @s animation.spartan_shield.layer_${layer}_index_${type}`);
		};
		editDynamicProperty(entity,"armor_stand_shield_layers",JSON.stringify(layers??[]));
	}
	resetSelectedSlot(){
			world.events.playerSpawn.subscribe((e)=>{
				this.selectedSlots.set(e.player,-1);
			});
	}
	netheriteMaterial(types){
		world.events.entitySpawn.subscribe(({entity})=>{
			if(entity?.typeId!="minecraft:item") return;
			let itemStack = entity.getComponent("minecraft:item").itemStack;
			if(!types.some((type)=>type==itemStack.typeId)) return;
			entity.triggerEvent("minecraft:netherite_material");
		});
	}
	interactArmorStandListener(){
		/*As you can see it is very limited*/
		world.events.beforeDataDrivenEntityTriggerEvent.subscribe(  (e)=>{
			if((e?.entity?.typeId!="minecraft:armor_stand")||(e.id!="minecraft:interact_listener")) return;
			let playerCollection = e.entity.dimension.getEntities({type: "minecraft:player", closest: 1});
			const [player] = playerCollection;
			if(!player) return;
		    this.printPatternsOnArmorStand(player.shieldLayers,e.entity);
		});
		world.events.beforeDataDrivenEntityTriggerEvent.subscribe(  (e)=>{
			if((e?.entity?.typeId!="minecraft:armor_stand")||(e.id!="minecraft:update_layers")) return;
			let layers = JSON.parse(fromDynamicProperty(e.entity,"armor_stand_shield_layers"));
		    this.printPatternsOnArmorStand(layers,e.entity);
		});
		world.events.entitySpawn.subscribe((e)=>{
			if(e.entity.typeId!="minecraft:armor_stand") return;
			var funcion = this.printPatternsOnArmorStand;
			setTimeOut(()=>{
				try{
					let layers = JSON.parse(fromDynamicProperty(e.entity,"armor_stand_shield_layers"));
					funcion(layers,e.entity);
				}catch(error){}
            },200);
		});
	}
	removeEffectAttacks(player){
		let badEffects = [ "poison", "wither", "hunger", "slowness", "levitation", "blindness" ];
		for(let effect of badEffects){
			let effectType = player.getEffect(MinecraftEffectTypes[effect]);
			if(effectType===undefined){
				setTimeOut(()=>{
					try{
					if(player.getEffect(MinecraftEffectTypes[effect])){
						player.removeTag(effect+"_from_attack");
						player.addEffect(MinecraftEffectTypes[effect],0, player.getEffect(MinecraftEffectTypes[effect]).amplifier+1);
					}}catch(error){};
                },6);
			}else{ let {duration, amplifier } = effectType; player.addEffect(MinecraftEffectTypes[effect],0, amplifier+1); setTimeOut(()=>{try{player.addEffect(MinecraftEffectTypes[effect],duration-1,amplifier)}catch(error){}},2);};
		};
	}
}
const n = new CustomShields(["add:wooden_shield", "add:golden_shield", "add:stone_shield", "add:iron_shield", "add:diamond_shield", "add:obsidian_shield", "add:netherite_shield"]);