import {world,TickEventSignal,Player,Block,DynamicPropertiesDefinition, MinecraftEntityTypes,Entity} from "@minecraft/server";
import {ActionFormResponse, ModalFormResponse, MessageFormResponse, ActionFormData, ModalFormData, MessageFormData} from "@minecraft/server-ui";
import {cloneObject,executeFormAction} from "./utils.js";


export class ParseDialogue {
	constructor(data){
        this.data = data;
        this.player = null;
        this.target = null;
        this.current_page = 0;
        this.dynamic_attributes = [ "title", "body", "name", "show", "value", "icon", "collection", "min", "max", "step", "placeHolder" ];
        this.element_data = new Map();
        
        this.ui_style = "default"
	}
	async display(player,target,manager){
		
		this.data.pages = this.data.pages.sort(function(a,b){ return a.page - b.page;});
		this.player = player; this.target = target;
		let run = true;
		try{
		while(run){
			this.clearData();
			let screen = this.data.pages[this.current_page];
			
			let form = this.parser(this.getType(screen.type),{ player: this.player, target: this.target, screen: screen});
			let screen_pending_name = manager.getPendingScreen(this.player);
			let r = await form.show(player);
		
			if(r instanceof ModalFormResponse){
				if(!r.canceled){
					
					for(let el=0; el<r.formValues.length;el++){
						let a = this.element_data.get(el);
						(typeof a?.action == "function") ? executeFormAction(a,{player:player, target:target, value:r.formValues[el], values: r.formValues, element: a}) : undefined;
						
					}
				}
			}
			if(r instanceof ActionFormResponse){
				if(!r.canceled){
					
					let a = this.element_data.get(r.selection);
					(typeof a?.action == "function") ? a?.action({player:player,target:target, value: r.selection}) : undefined;
					if((a?.to!=null)&&(Number.isInteger(a?.to))){ this.current_page = a.to; this.clearData(); continue; };
				}else{ try{screen.onClose({player: player, target: target});}catch(error){}}
			}
			if(r instanceof MessageFormResponse){
				
				if(!r.canceled){
					
					let a = this.element_data.get(r.selection);
					(typeof a?.action == "function") ? a?.action({player:player,target:target}) : undefined;
					if((a?.to!=null)&&(Number.isInteger(a?.to))){ this.current_page = a.to; this.clearData(); continue; };
				}
		    }
			run = false;
			(r.cancelationReason == "userBusy"&&(screen_pending_name == "none")&&(this.data.force==true)) ? manager.setPendingScreen(this.player,this.data.name) : undefined;
			((screen_pending_name==this.data.name)&&(r.cancelationReason!="userBusy")) ? manager.deletePendingScreen(this.player) : undefined;
		}}catch(error){run=false;console.warn("error 2",error); manager.onClose(player);};
		return void(manager.onClose(player));
	}
	parser(form, context){
		let { player, target, screen } = context;
		let page = cloneObject(screen);
		for(let attr in page){
			
			if(!this.dynamic_attributes.find((e)=>e==attr)) continue;
			try{
			(typeof page[attr] == "function") ? page[attr]=page[attr]({player:player,target:target}) : undefined;}catch(error){ console.warn("error ", error); };
		};
		for(let el of page.elements){
			for(let attr in el){
				if(!this.dynamic_attributes.find((e)=>e==attr)) continue;
				try{
			    ((typeof el[attr] == "function")) ? el[attr]=el[attr]({player:player,target:target}) : undefined;}catch(error){ console.warn("error en " + attr, error); };
			}
		}
		
		if(form instanceof ModalFormData){
			
			page.elements = page.elements.filter((i)=>{return i.show==true;});
			form.title(page.title);
			let index =0;
			
			for(let el =0; el<page.elements.length;el++){
				let element = page.elements[el];
				
				if(element.type!="icon"){
					switch(element.type){
			          case("toggle"):
			            
			            form.toggle(element.name,element.value??null);  break;
			          case("dropdown"):
			
			            form.dropdown(element.name,element.collection,element.value??null); break;
			          case("textField"):
			
			            form.textField(element.name,element.placeHolder,element.value??null); break;
			          case("slider"):
			
			            form.slider(element.name,element.min,element.max,element.step,element.value??null); break;
		            }
		            (typeof element.action == "function") ? this.element_data.set(index,element) : undefined;
		            index+=1;
				}
			}
		}
		if(form instanceof MessageFormData){

			page.elements.length = 2;
			form.title((page.title??"")).body((page.body??""));
			form.button1(page.elements[1].name);
			form.button2(page.elements[0].name);
			this.element_data.set(0,page.elements[0]);this.element_data.set(1,page.elements[1]);
			
		}
		if(form instanceof ActionFormData){
			page.elements = page.elements.filter((i)=>{return i.show==true;});
			form.title((page.title??"")); form.body((page.body));
			for(let el=0; el<page.elements.length;el++){
				form.button(page.elements[el].name,page.elements[el].icon??null);
				this.element_data.set(el,page.elements[el]);
			};
			(page.elements.length<=0) ? form.button("formSheets.actionFormData.noButtons") : undefined;
		}
		return form;
	}
	clearData(){
		this.element_data = new Map();
	}
	getType(name){
		switch(name){
			case("modal"):
			  return new ModalFormData();
			case("message"):
			  return new MessageFormData();
			case("action"):
			  return new ActionFormData();
		}
	}
};
export class ScreenEvents{
	static onPlayerPlace(responseCallback,screen){
		return {event: world.events.blockPlace, response: responseCallback, screen: screen };
	}
	static onPlayerDestroy(responseCallback, screen){
		return {event: world.events.blockDestroy, response: responseCallback, screen: screen};
	}
	static entityHit(responseCallback, screen){
		return {event: world.events.entityHit, response: responseCallback, screen: screen};
	}
	static beforeChat(responseCallback,screen){
		return {event: world.events.beforeChat, response: responseCallback, screen: screen};
	}
	static messageReceive(responseCallback,screen){
		return {event: world.events.messageReceive, response: responseCallback, screen: screen};
	}
	static tick(responseCallback,screen){
		return {event: world.events.tick, response: responseCallback, screen: screen};
	}
	static beforeItemUseOn(responseCallback, screen){
		return {event: world.events.beforeItemUseOn, response: responseCallback, screen: screen};
	}
}
/*ScreenEvents methods are world events that requires a callback that will be executed inside of this event, this callback must return boolean */
export class ScreenDisplay{
	constructor(){
		/*Identifier is the returned callBack event and its value is the same*/
		this.screenEvents = new Map();
		this.forcedScreens = new Map();
		/*Identifier is the player and its value is the event*/
		this.initialize = world.events.worldInitialize.subscribe((e)=>{ let d = new DynamicPropertiesDefinition(); d.defineBoolean("is_player_in_ui"); d.defineString("screen_pending", 100); e.propertyRegistry.registerEntityTypeDynamicProperties(d, MinecraftEntityTypes.player);});
		this.define = world.events.playerSpawn.subscribe((e)=>{ this.onClose(e.player); this.deletePendingScreen(e.player); });
		this.forced_screens = world.events.tick.subscribe((e)=>{
			const players = world.getPlayers();
			for(const player of players){
		
				for(const action of this.forcedScreens.values()){
					const info = action.response(player);
			        const is_in_ui = info.player?.getDynamicProperty("is_player_in_ui");
			        if((info.player instanceof Player)&&((info.target instanceof Entity)||(info.target instanceof Block)||(info.target instanceof Player))&&(!is_in_ui)) {
                        let pantalla = new ParseDialogue(action.screen);
                        this.onEnter(info.player);     
                        pantalla.display(info.player,info.target,this);
                   };
				}
			}
		});
	}
	register(screenEvent){
		if(screenEvent.event instanceof TickEventSignal){
			this.registerForcedScreen(screenEvent);
		}
		else{
		const event = screenEvent.event.subscribe((e)=>{
			if(!this.screenEvents.has(event)) return void(0);
			let {response, screen} = this.screenEvents.get(event);
			const info = response (e);
			const is_in_ui = info.player?.getDynamicProperty("is_player_in_ui");
			if((info.player instanceof Player)&&((info.target instanceof Entity)||(info.target instanceof Block)||(info.target instanceof Player))&&(!is_in_ui)) {
              let pantalla = new ParseDialogue(screen);
              this.onEnter(info.player);     
              pantalla.display(info.player,info.target,this);
              return;
            };
		});
		(this.screenEvents.has(event)) ? this.delete(screenEvent,event) : this.screenEvents.set(event,screenEvent);
		}
	}
	delete(screenEvent,event){
		screenEvent.event.unsubscribe(event);
	}
	onClose(player){
		try{
			player.setDynamicProperty("is_player_in_ui", false);
		}catch(error){};
	}
	onEnter(player){
		try{
			player.setDynamicProperty("is_player_in_ui", true);
		}catch(error){};
	}
	deletePendingScreen(player){
		try{
			player.setDynamicProperty("screen_pending","none");
		}catch(error){};
	}
	getPendingScreen(player){
		let screen_name = "none";
		try{
			let value = (player.getDynamicProperty("screen_pending")??"none");
			screen_name = value;
		}catch(error){ screen_name = "none" };
		return screen_name;
	}
	setPendingScreen(player,name){
		try{
			let value = this.getPendingScreen(player);
			(value=="none") ? player?.setDynamicProperty("screen_pending",name) : undefined;
		}catch(error){player?.setDynamicProperty("screen_pending","none"); };
	}
	registerForcedScreen(screenEvent){
		this.forcedScreens.set(screenEvent.response,screenEvent);
	}
}